#include "URLLockNative.h"
#include <nsXPCOM.h>
#include <nsIMemory.h>

/* Implementation file */
NS_IMPL_ISUPPORTS1(URLLockNative, IURLLockNative)

/**
 * Creates the native lock object
 */
URLLockNative::URLLockNative()
{
}

/**
 * Destructs the native object, ensuring that all PCRE
 * resources are freed in the process
 */
URLLockNative::~URLLockNative()
{
}

/**
 * string getUserFullName (in nsISupportsVoid userID);
 * Returns the full name of the given user
 */
NS_IMETHODIMP URLLockNative::GetUserFullName(nsISupportsVoid *userID, char **_retval)
{
	// Get the user from the input user ID
	PUSER_INFO_2 user;
	((UserInfoSupportsVoid*) userID)->GetData((void**) &user);

	// Return the user's full name
	*_retval = convertUnicodeToMozillaASCII(user->usri2_full_name);
    return NS_OK;
}

/**
 * void getUserIDs (out PRUint32 count, [array, size_is (count), retval] out nsISupportsVoid list);
 * Returns an array of user IDs, which are opaque data objects that are dependant on the
 * operating system. These might not be integers on some operating systems, including
 * Microsoft Windows
 */
NS_IMETHODIMP URLLockNative::GetUserIDs(PRUint32 *count, nsISupportsVoid ***list)
{
    PUSER_INFO_2 userBuffer;
	DWORD netEnumResult, resumeHandle = NULL, entriesRead, totalEntries, nextEntry = 0;
	DWORD allocation = 0, i;
	UserInfoSupportsVoid **retval = 0;
	UserInfoSupportsVoid *parent;

	// Enter the user enumeration loop
	do {

		// Enumerate the user list
		netEnumResult = NetUserEnum(NULL, 2 /* for USER_INFO_2 structures */,
			FILTER_NORMAL_ACCOUNT, (LPBYTE *) &userBuffer, FILTER_NORMAL_ACCOUNT,
			&entriesRead, &totalEntries, &resumeHandle);

		// Allocate a sufficient amount of memory for the list, or, if already
		// allocated, update the allocation
		if (!allocation) {
			retval = (UserInfoSupportsVoid**) NS_Alloc(sizeof(UserInfoSupportsVoid*) * entriesRead);
			if (!retval) {
				NetApiBufferFree(userBuffer);
				return NS_ERROR_OUT_OF_MEMORY;
			}
			allocation = entriesRead;
		} else if (allocation < nextEntry + entriesRead) {
			UserInfoSupportsVoid** oldAllocation = retval;
			retval = (UserInfoSupportsVoid**) NS_Realloc(retval, sizeof(UserInfoSupportsVoid*) * (nextEntry + entriesRead));
			if (!retval) {
				NetApiBufferFree(userBuffer);
				NS_Free(oldAllocation);
				return NS_ERROR_OUT_OF_MEMORY;
			}
			allocation = nextEntry + entriesRead;
		}

		// Nullify parent and lazily-create
		parent = NULL;

		// Process each user
		for (i = 0; i < entriesRead; i++) {

			// If this user is a normal user, then proceed
			if (userBuffer[i].usri2_flags & UF_NORMAL_ACCOUNT) {

				// Wrap the returned buffer lazily
				if (!parent) {
					parent = new UserInfoSupportsVoid(userBuffer);
				}

				// Create a UserInfoSupportsVoid wrapper for the PUSER_INFO_2 element
				// and store it into the return value
				retval[nextEntry] = new UserInfoSupportsVoid(parent, &userBuffer[i]);
				retval[nextEntry]->AddRef();

				// Increment nextEntry for the next entry
				nextEntry++;
			}
		}
	} while (netEnumResult == ERROR_MORE_DATA);

	// Set the return values
	*count = nextEntry;
	*list = (nsISupportsVoid**) retval;
	return NS_OK;
}

/**
 * Converts the input Unicode string to a Mozilla-allocated string. This
 * function returns null if memory could not be allocated or if the conversion
 * was not successful
 */
char* URLLockNative::convertUnicodeToMozillaASCII(LPCWSTR unicodeString)
{
	// Get the buffer size required for the conversion
	DWORD bufferSize = WideCharToMultiByte(CP_UTF8, 0, unicodeString, -1, NULL, 0, NULL, NULL);
	LPSTR buffer;
	if (!bufferSize) {
		return NULL;
	}

	// Allocate the buffer
	buffer = (LPSTR) NS_Alloc(bufferSize);

	// Convert the string to UTF-8
	if (!WideCharToMultiByte(CP_UTF8, 0, unicodeString, -1, buffer, bufferSize, NULL, NULL)) {
		NS_Free(buffer);
		return NULL;
	}

	// Return the converted string
	return buffer;
}

/**
 * string getUserName (in nsISupportsVoid userID);
 * Returns the short user name of the given user
 */
NS_IMETHODIMP URLLockNative::GetUserName(nsISupportsVoid *userID, char **_retval)
{
	// Get the user from the input user ID
	PUSER_INFO_2 user;
	((UserInfoSupportsVoid*) userID)->GetData((void**) &user);

	// Convert the user name to ASCII
	*_retval = convertUnicodeToMozillaASCII(user->usri2_name);
	return NS_OK;
}

/**
 * string getCurrentUser ();
 * Returns the user name of the current user
 */
NS_IMETHODIMP URLLockNative::GetCurrentUser(char **_retval)
{
	// Ask Windows for the current user's user name
	WKSTA_USER_INFO_0 *info;
	if (NetWkstaUserGetInfo(NULL, 0 /* for WKSTA_USER_INFO_0 */, (LPBYTE *) &info) != NERR_Success) {
		if (GetLastError() == ERROR_NOT_ENOUGH_MEMORY) {
			return NS_ERROR_OUT_OF_MEMORY;
		} else {
			return NS_ERROR_FAILURE;
		}
	}

	// Return the user name
	*_retval = convertUnicodeToMozillaASCII(info->wkui0_username);

	// Free the returned info structure
	NetApiBufferFree(info);
    return NS_OK;
}

/**
 * boolean isAdministrator (in nsISupportsVoid userID);
 * Returns true if the given user is an administrator or false if not
 */
NS_IMETHODIMP URLLockNative::IsAdministrator(nsISupportsVoid *userID, PRBool *_retval)
{
	// Get the user from the input user ID
	PUSER_INFO_2 user;
	((UserInfoSupportsVoid*) userID)->GetData((void**) &user);

	// Return whether or not this user is an administrator
	*_retval = user->usri2_priv == USER_PRIV_ADMIN;
    return NS_OK;
}

/**
 * string getPCREError ();
 * Returns the error message that got generated when parsing the regular expression.
 * This is null if no error occurred
 */
NS_IMETHODIMP URLLockNative::GetPCREError(char **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * long getPCREErrorOffset ();
 * Returns the character offset in the regular expression pattern that caused an error
 * when parsing the regular expression. This is -1 if no error occurred
 */
NS_IMETHODIMP URLLockNative::GetPCREErrorOffset(PRInt32 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * nsISupportsVoid pcreCompile (in string pattern);
 * Compiles the given pattern into a PATTERN opaque object. If unsuccessful, this
 * function returns false and sets pcreError and pcreErrorOffset
 */
NS_IMETHODIMP URLLockNative::PcreCompile(const char *pattern, nsISupportsVoid **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * boolean pcreExec (in nsISupportsVoid pattern, in nsISupportsVoid patternExtra, in string match);
 * Returns true if the given match string matches the given pattern. Otherwise, this function
 * returns false
 */
NS_IMETHODIMP URLLockNative::PcreExec(nsISupportsVoid *pattern, nsISupportsVoid *patternExtra, const char *match, PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * nsISupportsVoid pcreStudy (in nsISupportsVoid pattern);
 * Performs additional optimizations to the regular expression and returns a PATTERN_EXTRA
 * object that contains the results of the optimizations. That object can be null, and that
 * is okay. Pass the result of this to pcreExec as the patternExtra parameter, even if it
 * is null
 */
NS_IMETHODIMP URLLockNative::PcreStudy(nsISupportsVoid *pattern, nsISupportsVoid **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * string getFirefoxProfilePath (in string userRegistryHiveMountPoint);
 * Returns the full path of the user’s default Firefox profile folder
 */
NS_IMETHODIMP URLLockNative::GetFirefoxProfilePath(const char *userRegistryHiveMountPoint, char **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * string getInternetExplorerFavoritesPath (in string userRegistryHiveMountPoint);
 * Returns the full path of the user’s Internet Explorer “favorites” path, which
 * are analogous to Firefox bookmarks
 */
NS_IMETHODIMP URLLockNative::GetInternetExplorerFavoritesPath(const char *userRegistryHiveMountPoint, char **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * IURLLockBrowserLinkList getInternetExplorerHistory ();
 * Returns the Internet Explorer history for the current user
 */
NS_IMETHODIMP URLLockNative::GetInternetExplorerHistory(IURLLockBrowserLinkList **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * string getUserRegistryHiveFileName (in nsISupportsVoid userID);
 * Return the file name of the given user’s registry hive
 */
NS_IMETHODIMP URLLockNative::GetUserRegistryHiveFileName(nsISupportsVoid *userID, char **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * boolean loadUserRegistryHive (in string fileName, in string mountPoint);
 * Loads the given registry hive file into the given registry mount point. If this was
 * not successful, an exception is thrown
 */
NS_IMETHODIMP URLLockNative::LoadUserRegistryHive(const char *fileName, const char *mountPoint, PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * void setUserRegistryWriteEnabled (in string key, in string user, in boolean enabled);
 * Sets whether or not the given key should permit write access to the given user
 */
NS_IMETHODIMP URLLockNative::SetUserRegistryWriteEnabled(const char *key, const char *user, PRBool enabled)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/**
 * boolean unloadUserRegistryHive (in string mountPoint);
 * Unloads the mounted registry hive at the given registry mount point. If this was not
 * successful, an exception is thrown
 */
NS_IMETHODIMP URLLockNative::UnloadUserRegistryHive(const char *mountPoint, PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}
